@@ -1,5 +1,15 @@
 #include "dbd_oracle.h"
 
+#define ORA_US7ASCII_CSID 1
+#define ORA_UTF8_CSID 871
+#define ORA_AL32UTF8_CSID 873
+#define ORA_UTF16_CSID 2000
+
+#define CS_IS_UTF8( x ) ( ( (x) == ORA_UTF8_CSID ) || ( (x) == ORA_AL32UTF8_CSID ) )
+
+#define CSFORM_IMPLIED_CSID( x,y ) ( ( (x) == SQLCS_NCHAR ) ? y->ncharsetid : y->charsetid )
+
+#define CSFORM_IMPLIES_UTF8( x,y ) ( CS_IS_UTF8( CSFORM_IMPLIED_CSID ( x,y ) ) )
 /*
  * Converts SQLite types to Lua types
  */
@@ -33,26 +43,49 @@ static void statement_fetch_metadata(lua_State *L, statement_t *statement) {
     int i;
 
     char errbuf[100];
-    int errcode;
+    sb4 errcode;
     int rc;
 
     if (statement->metadata)
 	return;
+    
+#if 0
+    // staged for later, when we'll pre-check the attribute sizes and
+    // types in order to set up the prefetch cache more intelligently
+    rc = OCIStmtExecute(
+			statement->conn->svc,
+			statement->stmt,
+			statement->conn->err,
+			(ub4)0, // type == OCI_STMT_SELECT ? 0 : 1,
+			(ub4)0,
+			(CONST OCISnapshot *)NULL,
+			(OCISnapshot *)NULL,
+			OCI_DESCRIBE_ONLY
+			);
 
-    ub4 prefetch_mem = 1024*1024;
-    if (OCIAttrSet(statement->stmt, OCI_HTYPE_STMT, &prefetch_mem, sizeof(prefetch_mem),
-		   OCI_ATTR_PREFETCH_MEMORY, statement->conn->err) != OCI_SUCCESS) {
-      OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
-      luaL_error(L, "prefetch_mem set %s", errbuf);
+    if (rc) {
+      OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+      luaL_error(L, "describe %s", errbuf);
+    }
+#endif
+    
+    ub4 prefetch_mem = statement->prefetch_mem;
+    if (prefetch_mem) {
+      if (OCIAttrSet(statement->stmt, OCI_HTYPE_STMT, &prefetch_mem, sizeof(prefetch_mem),
+		     OCI_ATTR_PREFETCH_MEMORY, statement->conn->err) != OCI_SUCCESS) {
+	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *)errbuf, (ub4) sizeof(errbuf), (ub4)OCI_HTYPE_ERROR);
+	luaL_error(L, "prefetch_mem set %s", errbuf);
+      }
     }
 
-    // FIXME: estimate this based on the prefetch mem size and our query's description
-    ub4 prefetch_rows = 1000000;
-    if (OCIAttrSet(statement->stmt, OCI_HTYPE_STMT,
-		   &prefetch_rows, sizeof(prefetch_rows), OCI_ATTR_PREFETCH_ROWS,
-		   statement->conn->err) != OCI_SUCCESS) {
-      OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
-      luaL_error(L, "prefetch_rows set %s", errbuf);
+    ub4 prefetch_rows = statement->prefetch_rows;
+    if (prefetch_rows) {
+      if (OCIAttrSet(statement->stmt, OCI_HTYPE_STMT,
+		     &prefetch_rows, sizeof(prefetch_rows), OCI_ATTR_PREFETCH_ROWS,
+		     statement->conn->err) != OCI_SUCCESS) {
+	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *)errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR);
+	luaL_error(L, "prefetch_rows set %s", errbuf);
+      }
     }
 
     statement->bind = (bindparams_t *)malloc(sizeof(bindparams_t) * statement->num_columns);
@@ -62,45 +95,95 @@ static void statement_fetch_metadata(lua_State *L, statement_t *statement) {
     for (i = 0; i < statement->num_columns; i++) {
 	rc = OCIParamGet(statement->stmt, OCI_HTYPE_STMT, statement->conn->err, (dvoid **)&bind[i].param, i+1);
 	if (rc) {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	  OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR);
 	    luaL_error(L, "param get %s", errbuf);
 	}
 
 	rc = OCIAttrGet(bind[i].param, OCI_DTYPE_PARAM, (dvoid *)&(bind[i].name), (ub4 *)&(bind[i].name_len), OCI_ATTR_NAME, statement->conn->err);
 	if (rc) {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	  OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), (ub4) OCI_HTYPE_ERROR);
 	    luaL_error(L, "name get %s", errbuf);
 	}
 
 	rc = OCIAttrGet(bind[i].param, OCI_DTYPE_PARAM, (dvoid *)&(bind[i].data_type), (ub4 *)0, OCI_ATTR_DATA_TYPE, statement->conn->err);
 	if (rc) {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 	    luaL_error(L, "datatype get %s", errbuf);
 	}
 
 	rc = OCIAttrGet(bind[i].param, OCI_DTYPE_PARAM, (dvoid *)&(bind[i].max_len), 0, OCI_ATTR_DATA_SIZE, statement->conn->err);
 	if (rc) {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 	    luaL_error(L, "datasize get %s", errbuf);
 	}
 
-	/* TO_CHAR conversion wants one more byte than it actually reports? :/
-	 *  Try this:  local sth = assert(db:prepare("select to_char(sysdate, 'YYYYMMDD') FROM dual"))
-	 *             assert(sth:execute())
-	 * It reports max_len==8, but then 
-	 *
-         *   Fetch failed ORA-01406: fetched column value was truncated
-	 *
-	 * ... unless we add one more byte. This seems like a general
-	 * problem waiting to be more of a nuisance.
-         */
+#ifdef OCI_ATTR_CHARSET_ID
+	rc = OCIAttrGet(bind[i].param, OCI_DTYPE_PARAM, (dvoid *)&bind[i].csid, 0, OCI_ATTR_CHARSET_ID, statement->conn->err);
+	rc = OCIAttrGet(bind[i].param, OCI_DTYPE_PARAM, (dvoid *)&bind[i].csform, 0, OCI_ATTR_CHARSET_FORM, statement->conn->err);
+#endif
 
-	bind[i].max_len++;
+	switch (bind[i].data_type) {
+	case SQLT_CHR: // VARCHAR2 ("character string")
+	  // sometimes o11 returns 0 as the length. Substitute an
+	  // arbitrary default.
+	  if (bind[i].max_len == 0) {
+	    bind[i].max_len = 4000;
+	  }
+	  /* Fallthrough to UTF8 handling... */
+	  
+	case SQLT_AFC: // CHAR ("ANSI fixed character")
+	  /* If the field's charset isn't UTF-8, but the session
+	   * environment implies UTF-8, then we have to account for
+	   * the extra required space manually.
+	   */
+#ifdef OCI_ATTR_CHARSET_ID
+	  if (!CS_IS_UTF8(bind[i].csid) && CSFORM_IMPLIES_UTF8(bind[i].csform, statement->conn) )
+#endif
+	    bind[i].max_len *= 4;
+	  break;
+
+	case SQLT_TIMESTAMP:
+	case SQLT_TIMESTAMP_TZ:
+	case SQLT_TIMESTAMP_LTZ:
+	  // size depends on the NLS default date format. Assume a generous default.
+	  bind[i].max_len = 200;
+	  break;
+	  
+	default:
+	  // Unhandled type...
+	  //	  printf("type %d [%d]\n", bind[i].data_type, bind[i].max_len);
+	  break;
+	}
+	
+	if (bind[i].data_type == SQLT_BLOB ||
+	    bind[i].data_type == SQLT_CLOB) {
+	  // BLOB and CLOB types are segregated because we want to
+	  // retrieve them as different types.  This also sets us up
+	  // to be able to call this with OCI_DYNAMIC_FETCH later:
+	  // retrieving the LOB in chunks.
+	  bind[i].data = calloc(bind[i].max_len, sizeof(char));
+	  rc = OCIDefineByPos(statement->stmt,
+			      &bind[i].define,
+			      statement->conn->err,
+			      (ub4)i+1,
+			      bind[i].data,
+			      bind[i].max_len,
+			      bind[i].data_type == SQLT_BLOB ? SQLT_BIN : SQLT_CHR,
+			      (dvoid *)&(bind[i].null),
+			      (ub2 *)&(bind[i].ret_len),
+			      (ub2 *)&(bind[i].ret_err),
+			      (ub4)OCI_DEFAULT);
+	} else {
+	  // This is a SQLT_STR return value, which needs space for a
+	  // NULL terminating character.  Add 4 bytes in case it's
+	  // UTF-8.
+	  bind[i].max_len+=4;
+	  bind[i].data = calloc(bind[i].max_len, sizeof(char));
+	  rc = OCIDefineByPos(statement->stmt, &bind[i].define, statement->conn->err, (ub4)i+1, bind[i].data, bind[i].max_len, SQLT_STR, (dvoid *)&(bind[i].null), (ub2 *)&(bind[i].ret_len), (ub2 *)&(bind[i].ret_err), (ub4)OCI_DEFAULT);
+	}
 
-	bind[i].data = calloc(bind[i].max_len+1, sizeof(char));
-	rc = OCIDefineByPos(statement->stmt, &bind[i].define, statement->conn->err, (ub4)i+1, bind[i].data, bind[i].max_len, SQLT_STR, (dvoid *)&(bind[i].null), (ub2 *)0, (ub2 *)0, (ub4)OCI_DEFAULT);
 	if (rc) {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4)sizeof(errbuf), OCI_HTYPE_ERROR);
+	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4)sizeof(errbuf), OCI_HTYPE_ERROR);
 	    luaL_error(L, "define by pos %s", errbuf);
 	}
     }
@@ -115,7 +198,6 @@ static void statement_fetch_metadata(lua_State *L, statement_t *statement) {
 static int statement_affected(lua_State *L) {
     statement_t *statement = (statement_t *)luaL_checkudata(L, 1, DBD_ORACLE_STATEMENT);
     int affected;
-    int rc; 
 
     if (!statement->stmt) {
         luaL_error(L, DBI_ERR_INVALID_STATEMENT);
@@ -124,7 +206,7 @@ static int statement_affected(lua_State *L) {
     /* 
      * get number of affected rows 
      */
-    rc = OCIAttrGet(
+    OCIAttrGet(
 	(dvoid *)statement->stmt, 
 	(ub4)OCI_HTYPE_STMT,
         (dvoid *)&affected, 
@@ -146,9 +228,7 @@ int statement_close(lua_State *L) {
     int ok = 0;
 
     if (statement->stmt) {
-	int rc;
-
-	rc = OCIHandleFree((dvoid *)statement->stmt, OCI_HTYPE_STMT);    /* Free handles */	
+        OCIHandleFree((dvoid *)statement->stmt, OCI_HTYPE_STMT);    /* Free handles */	
 
 	statement->stmt = NULL;
     }
@@ -180,7 +260,7 @@ static int statement_columns(lua_State *L) {
 
     lua_newtable(L);
     for (i = 0; i < statement->num_columns; i++) {
-	const char *name = dbd_strlower(statement->bind[i].name);
+        const char *name = dbd_strlower((char *)statement->bind[i].name);
 
 	LUA_PUSH_ARRAY_STRING(d, name);
     }
@@ -202,7 +282,7 @@ int statement_execute(lua_State *L) {
     int rc;
 
     char errbuf[100];
-    int errcode;
+    sb4 errcode;
 
     ub2 type;
 
@@ -211,12 +291,12 @@ int statement_execute(lua_State *L) {
 	lua_pushstring(L, DBI_ERR_EXECUTE_INVALID);
 	return 2;
     }
-
     for (p = 2; p <= n; p++) {
 	int i = p - 1;
 	int type = lua_type(L, p);
 	char err[64];
 	const char *value;
+	size_t val_size;
 
 	OCIBind *bnd = (OCIBind *)0;
 
@@ -240,16 +320,16 @@ int statement_execute(lua_State *L) {
 	case LUA_TNUMBER:
 	case LUA_TSTRING:
 	case LUA_TBOOLEAN:
-	    value = lua_tostring(L, p);
+	  value = lua_tolstring(L, p, &val_size);
 
 	    errflag = OCIBindByPos(
 		statement->stmt, 
 		&bnd, 
 		statement->conn->err, 
-		i, 
-		value, 
-		strlen(value), 
-		SQLT_CHR, 
+		(ub4)i, 
+		(dvoid *)value, 
+		(sb4)val_size,
+		(ub2)SQLT_CHR, 
 		(dvoid *)0, 
 		(ub2 *)0, 
 		(ub2 *)0, 
@@ -275,7 +355,7 @@ int statement_execute(lua_State *L) {
 	if (errstr)
 	    lua_pushfstring(L, DBI_ERR_BINDING_PARAMS, errstr);
 	else {
-	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	    OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 
 	    lua_pushfstring(L, DBI_ERR_BINDING_PARAMS, errbuf);
 	}
@@ -296,7 +376,7 @@ int statement_execute(lua_State *L) {
     );
 
     if (rc) {
-	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 
 	lua_pushboolean(L, 0);
 	lua_pushfstring(L, "Error getting type: %s", errbuf);
@@ -319,10 +399,10 @@ int statement_execute(lua_State *L) {
     );
 
     if (rc) {
-	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 
 	lua_pushboolean(L, 0);
-	lua_pushfstring(L, DBI_ERR_BINDING_PARAMS, errbuf);
+	lua_pushfstring(L, DBI_ERR_BINDING_EXEC, errbuf);
 
 	return 2;
     }
@@ -340,7 +420,7 @@ int statement_execute(lua_State *L) {
     );
 
     if (rc) {
-	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, &errcode, errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
+	OCIErrorGet((dvoid *)statement->conn->err, (ub4) 1, (text *) NULL, (sb4 *)&errcode, (text *) errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
 
 	lua_pushboolean(L, 0);
 	lua_pushfstring(L, DBI_ERR_BINDING_PARAMS, errbuf);
@@ -362,14 +442,14 @@ static int statement_fetch_impl(lua_State *L, statement_t *statement, int named_
     bindparams_t *bind;
 
     char errbuf[100];
-    int errcode;
+    sb4 errcode;
 
     if (!statement->stmt) {
 	luaL_error(L, DBI_ERR_FETCH_INVALID);
 	return 0;
     }
 
-    statement_fetch_metadata(L, statement);    
+    statement_fetch_metadata(L, statement);
     bind = statement->bind;
 
     status = OCIStmtFetch(statement->stmt, statement->conn->err, 1, OCI_FETCH_NEXT, OCI_DEFAULT);
@@ -378,11 +458,10 @@ static int statement_fetch_impl(lua_State *L, statement_t *statement, int named_
 	/* No more rows */
         lua_pushnil(L);
         return 1;
-    } else if (status != OCI_SUCCESS) {
-	OCIErrorGet((dvoid *)statement->conn->err, (ub4)1, (text *)NULL, &errcode, errbuf, (ub4)sizeof(errbuf), OCI_HTYPE_ERROR);
-	luaL_error(L, DBI_ERR_FETCH_FAILED, errbuf);
     }
 
+    // Loop through the fields, even on error; the error might be 1406 (truncated column), and we might want to return partial results...
+    
     if (statement->num_columns) {
 	int i;
 	int d = 1;
@@ -391,9 +470,23 @@ static int statement_fetch_impl(lua_State *L, statement_t *statement, int named_
 
 	for (i = 0; i < statement->num_columns; i++) {
 	    lua_push_type_t lua_push = oracle_to_lua_push(bind[i].data_type, bind[i].null);
-	    const char *name = dbd_strlower(bind[i].name);
+	    const char *name = dbd_strlower((char *)bind[i].name);
 	    const char *data = bind[i].data;
+	    size_t data_size = bind[i].ret_len;
 
+	    if ((bind[i].data_type == SQLT_BLOB ||
+		 bind[i].data_type == SQLT_CLOB) &&
+		status == 1 &&
+		bind[i].ret_err == 1406) {
+	      // Allow partial return from a LOB
+	      status = 0;
+	    }
+	    else if (bind[i].ret_err != 0 &&
+		     !(bind[i].ret_err == 1405 && bind[i].null)) {
+	      // If we need debugging...
+	      // printf("Error %d with column %.*s\n", bind[i].ret_err, bind[i].name_len, bind[i].name);
+	    }
+		     
 	    if (lua_push == LUA_PUSH_NIL) {
                 if (named_columns) {
                     LUA_PUSH_ATTRIB_NIL(name);
@@ -418,9 +511,9 @@ static int statement_fetch_impl(lua_State *L, statement_t *statement, int named_
                 }
             } else if (lua_push == LUA_PUSH_STRING) {
                 if (named_columns) {
-                    LUA_PUSH_ATTRIB_STRING(name, data);
+		    LUA_PUSH_ATTRIB_STRING_BY_LENGTH(name, data, data_size);
                 } else {
-                    LUA_PUSH_ARRAY_STRING(d, data);
+		    LUA_PUSH_ARRAY_STRING_BY_LENGTH(d, data, data_size);
                 }
             } else if (lua_push == LUA_PUSH_BOOLEAN) {
 		int val = atoi(data);
@@ -441,6 +534,11 @@ static int statement_fetch_impl(lua_State *L, statement_t *statement, int named_
 	lua_pushnil(L);
     }
 
+    if (status != OCI_SUCCESS) {
+      OCIErrorGet((dvoid *)statement->conn->err, (ub4)1, (text *)NULL, (sb4 *)&errcode, (text *) errbuf, (ub4)sizeof(errbuf), OCI_HTYPE_ERROR);
+      luaL_error(L, DBI_ERR_FETCH_FAILED, errbuf);
+    }
+
     return 1;    
 }
 
@@ -508,7 +606,6 @@ static int statement_tostring(lua_State *L) {
 }
 
 int dbd_oracle_statement_create(lua_State *L, connection_t *conn, const char *sql_query) { 
-    int rc;
     statement_t *statement = NULL;
     OCIStmt *stmt;
     char *new_sql;
@@ -518,8 +615,8 @@ int dbd_oracle_statement_create(lua_State *L, connection_t *conn, const char *sq
      */
     new_sql = dbd_replace_placeholders(L, ':', sql_query);
 
-    rc = OCIHandleAlloc((dvoid *)conn->oracle, (dvoid **)&stmt, OCI_HTYPE_STMT, 0, (dvoid **)0);
-    rc = OCIStmtPrepare(stmt, conn->err, new_sql, strlen(new_sql), (ub4)OCI_NTV_SYNTAX, (ub4)OCI_DEFAULT);
+    OCIHandleAlloc((dvoid *)conn->oracle, (dvoid **)&stmt, OCI_HTYPE_STMT, 0, (dvoid **)0);
+    OCIStmtPrepare(stmt, conn->err, (CONST text *)new_sql, (ub4)strlen(new_sql), (ub4)OCI_NTV_SYNTAX, (ub4)OCI_DEFAULT);
 
     free(new_sql);
 
@@ -529,6 +626,8 @@ int dbd_oracle_statement_create(lua_State *L, connection_t *conn, const char *sq
     statement->num_columns = 0;
     statement->bind = NULL;
     statement->metadata = 0;
+    statement->prefetch_mem = conn->prefetch_mem;
+    statement->prefetch_rows = conn->prefetch_rows;
 
     luaL_getmetatable(L, DBD_ORACLE_STATEMENT);
     lua_setmetatable(L, -2);
