@@ -45,7 +45,7 @@ static int deallocate(statement_t *statement) {
 	 * - either by a mistake by the calling Lua program, or by
 	 * garbage collection. Don't die in that case.
 	 */
-	if (!statement->conn->postgresql) {
+	if (statement->conn->postgresql) {
 		snprintf(command, IDLEN+13, "DEALLOCATE \"%s\"", statement->name);    
     	result = PQexec(statement->conn->postgresql, command);
 
@@ -218,7 +218,12 @@ cleanup:
         lua_pushfstring(L, DBI_ERR_BINDING_EXEC, PQresultErrorMessage(result));
         return 2;
     }
-    
+
+    if (statement->result) {
+        status = PQresultStatus (statement->result);
+	if (status == PGRES_COMMAND_OK || status == PGRES_TUPLES_OK)
+            PQclear (statement->result);
+    }
     statement->result = result;
 
     lua_pushboolean(L, 1);
